//Gestionnaire de fiches Layered.
//------------------------------------------------------------------------------
//Dans la fiche concerne, il suffit de dclarer une variable de type
//TLayeredImage et de la crer dans son constructeur ou vnement OnCreate.

unit LayeredImage;

interface

uses
  Windows, SysUtils, Classes, Graphics, Forms, PNGImage, ContNrs, Math;

type
//------------------------------------------------------------------------------
  //Dfinition d'un pixel dans un bitmap 32 bits. (RGBA)
  PRGBQuadArray = ^TRGBQuadArray;
  TRGBQuadArray = array[0..0] of TRGBQuad;

//------------------------------------------------------------------------------
  TErase = (erNone, erAll, erDest);

//------------------------------------------------------------------------------
  //Classe permettant la gestion d'une fiche en mode layered.
  TLayeredImage = class
  private
    FForm          :TForm;             //La fiche rfrence
    FBitmap        :TBitmap;           //Image de fond de base. Est par exemple
                                       //utilise pour effacer une partie de
                                       //la fiche (InternalBmp)
    FInternalBmp   :TBitmap;           //Image finale vue par l'utilisateur
    FBmpStack      :TStack;            //Pile de type FILO d'images de fond
    FSize          :TPoint;            //Taille de la fiche
    FAlpha         :byte;              //Transparence de la fiche

  protected
    Blend :TBlendFunction;             //Mode de rendu de la fiche

  public
    property InternalBmp   :TBitmap read FInternalBmp;
    property Bitmap        :TBitmap read FBitmap;
    property Size          :TPoint  read FSize;
    property Alpha         :byte    read FAlpha write FAlpha;

    //Sauvegarde l'image de fond actuelle sur la pile
    procedure Push;

    //Rcupre la dernire image de fond mise sur la pile
    procedure Pop;

    //Rcupre la dernire image de fond sans la supprimer de la pile
    procedure Peek;

    //Dessine sur l'image.
    //Deux mthodes overload pour soit dessiner une portion prcise de l'image
    //source  un endroit donn ou sur toute la fiche.
    procedure Draw(const aSource :TBitmap; const aSourceRect, aDestRect :TRect; const aErase :TErase = erNone; const aOpacity :byte = 255); overload;
    procedure Draw(const aSource :TBitmap; const aErase :TErase = erNone; const aOpacity :byte = 255); overload;

    //Efface (par recopie du FBitmap) une portion de la fiche ou toute
    //la fiche.
    procedure Erase(const aFlush :boolean = FALSE); overload;
    procedure Erase(aRect :TRect; const aFlush :boolean = FALSE); overload;

    //Applique effectivement les dessins  l'cran.
    procedure Flush;

    //Assigne le bitmap de fond et rafraichi l'cran
    procedure SetBitmap(aBitmap :TBitmap);

    //No comment!
    constructor Create(const aForm :TForm; const aBkBmp :TBitmap = nil; const aFlush :boolean = TRUE); virtual;
    destructor  Destroy; override;
  end;

//------------------------------------------------------------------------------
//Routine pour la cration d'un bitmap 32bit.
function CreateBmp32(const aWidth, aHeight :integer) :TBitmap;

//Charge une image PNG et renvoi un bitmap 32 bits.
function LoadPngToBmp(const aName :string) :TBitmap;

implementation

//------------------------------------------------------------------------------
function CreateBmp32(const aWidth, aHeight :integer) :TBitmap;
var
  y :integer;
begin
  //Cration du bitmap rsultant
  Result := TBitmap.Create;

  //Initialisation
  with Result do
  begin
    PixelFormat := pf32bit;
    Width       := aWidth;
    Height      := aHeight;

    //Reset
    for y := 0 to Height -1 do
      ZeroMemory(Scanline[y], Width *SizeOf(dword));
  end;
end;

//------------------------------------------------------------------------------
function LoadPngToBmp(const aName :string) :TBitmap;
var
  Png  :TPngObject;
  RGBA :PRGBQuadArray;
  x    :integer;
  y    :integer;
begin
  Assert(FindResource(hInstance, PChar(aName), RT_RCDATA) <> 0, Format('LoadPngToBmp: La ressource %s n''existe pas', [aName]));

  //Cration de l'objet PNG
  Png := TPngObject.Create;

  try
    //Chargement du png
    Png.LoadFromResourceName(hInstance, aName);

    //Cration du bitmap rsultant
    with Png do
      Result := CreateBmp32(Width, Height);

    with Result do
    begin
      //Dessine le PNG sur le bitmap. Ne fonctionne qu'en 24 bits !!!
      Canvas.Draw(0, 0, Png); 

      //Copie de la couche alpha
      for y := 0 to Height -1 do
      begin
        RGBA := Scanline[y];

        for x := 0 to Width -1 do
          RGBA[x].rgbReserved := Png.AlphaScanline[y][x];
      end;
    end;

  finally
    Png.Free;
  end;
end;

{ TLayeredImage }

//------------------------------------------------------------------------------
procedure TLayeredImage.Erase(aRect: TRect; const aFlush: boolean);
var
  y     :integer;
  x     :integer;
  Width :integer;
  RGBA  :PRGBQuadArray;
begin
  //Evite que aRect dpasse les limites de la fiche (sinon AV)
  IntersectRect(aRect, aRect, FForm.ClientRect);

  //Efface une partie de l'image interne
  x     := aRect.Left;
  Width := aRect.Right -aRect.Left;

  for y := aRect.Top to aRect.Bottom -1 do
  begin
    RGBA := FInternalBmp.Scanline[y];
    ZeroMemory(@RGBA[x], Width *SizeOf(DWord));
  end;

  //Et redessine la partie depuis l'image de fond
  if Assigned(FBitmap) then
    Draw(FBitmap, aRect, aRect);

  //Applique sur l'cran
  if aFlush then Flush;
end;

//------------------------------------------------------------------------------
procedure TLayeredImage.Erase(const aFlush: boolean);
var
  R :TRect;
begin
  //Efface la fiche complte
  R := Rect(0, 0, FSize.X, FSize.Y);
  Erase(R, aFlush);
end;

//------------------------------------------------------------------------------
procedure TLayeredImage.Draw(const aSource: TBitmap; const aSourceRect,
          aDestRect: TRect; const aErase :TErase; const aOpacity :byte);
begin
  //Eface une partie de l'image interne
  case aErase of
    erAll  : Erase(Rect(0, 0, FInternalBmp.Width, FInternalBmp.Height));
    erDest : Erase(aDestRect);
  end;
    
  //Dessine sur le bitmap interne
  if Assigned(aSource) then
  begin
    //Fixe temporairement la transparence  celle souhaite
    Blend.SourceConstantAlpha := aOpacity;

    //Dessin
    AlphaBlend(FInternalBmp.Canvas.Handle,
               aDestRect.Left,
               aDestRect.Top,
               aDestRect.Right  -aDestRect.Left,
               aDestRect.Bottom -aDestRect.Top,
               aSource.Canvas.Handle,
               aSourceRect.Left,
               aSourceRect.Top,
               aSourceRect.Right  -aSourceRect.Left,
               aSourceRect.Bottom -aSourceRect.Top,
               Blend);

    //Remet la transparence globale
    Blend.SourceConstantAlpha := FAlpha;
  end;
end;

//------------------------------------------------------------------------------
procedure TLayeredImage.Draw(const aSource: TBitmap; const aErase :TErase; const aOpacity: byte);
var
  R :TRect;
begin
  //Dessine  partir du point 0,0 de la fiche
  R := Rect(0, 0, aSource.Width, aSource.Height);
  Draw(aSource, R, R, aErase, aOpacity);
end;

//------------------------------------------------------------------------------
procedure TLayeredImage.Flush;
const
  BmpTopLeft : TPoint = (x:0; y:0);
begin
  //Applique l'image interne sur la fiche
  UpdateLayeredWindow(FForm.Handle, 0, nil, @Size, FInternalBmp.Canvas.handle, @BmpTopLeft, 0, @Blend, ULW_ALPHA);
end;

//------------------------------------------------------------------------------
procedure TLayeredImage.Push;
var
  Bmp :TBitmap;
begin
  //Cre un nouveau bitmap
  Bmp := TBitmap.Create;

  //Copie le fond
  Bmp.Assign(FBitmap);

  //Le sauvegarde sur la pile
  FBmpStack.Push(Bmp);

  //Met l'image interne en fond
  FBitmap.Assign(FInternalBmp);
end;

//------------------------------------------------------------------------------
procedure TLayeredImage.Peek;
begin
  Assert(FBmpStack.Count > 0, 'TLayeredImage.Peek : Pas d''image dans le stack.');

  //Rcupre le dernier bitmap sans le supprimer de la pile
  FBitmap.Assign(FBmpStack.Peek);
end;

procedure TLayeredImage.Pop;
var
  Bmp :TBitmap;
begin
  Assert(FBmpStack.Count > 0, 'TLayeredImage.Pop : Pas d''image dans le stack.');

  //Rcupre le dernier de la pile...
  Bmp := FBmpStack.Pop;
  FBitmap.Assign(Bmp);

  Bmp.Free;
end;

//------------------------------------------------------------------------------
procedure TLayeredImage.SetBitmap(aBitmap: TBitmap);
begin
  FBitmap.Assign(aBitmap);
  Erase(TRUE);
end;

//------------------------------------------------------------------------------
destructor TLayeredImage.Destroy;
begin
  //Libre les bitmaps internes
  FBitmap.Free;
  FInternalBmp.Free;

  //Vide la pile et la dtruit
  while FBmpStack.Count > 0 do
    TBitmap(FBmpStack.Pop).Free;

  FBmpStack.Free;

  inherited;
end;

//------------------------------------------------------------------------------
constructor TLayeredImage.Create(const aForm :TForm; const aBkBmp :TBitmap; const aFlush :boolean);
begin
  Assert(Assigned(aForm), 'TLayeredImage.Create : La fiche ne peut tre nil');

  FForm   := aForm;
  FSize.X := FForm.Width;
  FSize.Y := FForm.Height;
  FAlpha  := 255;

  //Change le mode d'affichage de la fiche
  SetWindowLong(FForm.Handle, GWL_EXSTYLE, GetWindowLong(FForm.Handle, GWL_EXSTYLE) or WS_EX_LAYERED);

  //Cration des bitmaps internes 32 bits
  FBitmap      := CreateBmp32(Size.X, Size.Y);
  FInternalBmp := CreateBmp32(Size.X, Size.Y);

  //Cration de la pile
  FBmpStack := TStack.Create;

  //Initialisation du rendu
  with Blend do
  begin
    BlendOp             := AC_SRC_OVER;
    BlendFlags          := 0;
    SourceConstantAlpha := 255;
    AlphaFormat         := AC_SRC_ALPHA;
  end;

  Erase(aFlush);
end;

end.
