//Andnotor, developpez.net Defi N5 Sudoku solver
//------------------------------------------------------------------------------
// Unit sudoku solver
//------------------------------------------------------------------------------
unit uSudoku;

interface

uses Windows;

type
  //----------------------------------------------------------------------------
  // Dfinition de la grille

  //Cellule
  PSudokuCell = ^TSudokuCell;
  TSudokuCell = record
    Value  :byte;                       //Chiffre slectionn
    Row    :byte;                       //Index de ligne    (statique)
    Col    :byte;                       //Index de colonne  (statique)
    Region :byte;                       //Index de rgion   (statique)
    Mask   :word;                       //Masque des valeurs possibles
  end;

  //Masques possibles pour chaque bloc (lignes, colonnes et rgions)
  TSudokuBloc     = array[1..9] of word;

  //La grille (Reprsentation en ligne)
  TSudokuGrid     = array[1..81] of TSudokuCell;
  
  //Permet de lire la grille en ligne/colonne par transtypage
  //Accs plus facile depuis un TStringGrid
  TSudokuGridRC   = array[0..8, 0..8] of TSudokuCell;

  //Grille de travail. Pointeurs sur les cellules de la grille
  TSudokuWorkGrid = array[1..81] of PSudokuCell;


  //----------------------------------------------------------------------------
  //Classe TSudoku
  TSudoku = class
  private
    FGrid     : TSudokuGrid;            //Grille
    FDuration : extended;               //Dure de rsolution
    FSetValueCount : integer;           //Nombre de fois o une valeur 
                                        //t sette

  protected
    WorkGrid  : TSudokuWorkGrid;        //Grille de travail. Contient un
    Rows      : TSudokuBloc;            //Masques possibles pour les lignes
    Cols      : TSudokuBloc;            //Masques possibles pour les Colonnes
    Regions   : TSudokuBloc;            //Masques possibles pour les Rgions

    //Cherche la prochaine valeur possible pour une cellule
    function  NextValue(const aIndex :byte; var aValue :byte) :boolean;

    //Ecrit une valeur dans une cellule. Renvoi FALSE si valeur invalide.
    procedure InternalSetValue(const aIndex, aValue :byte); 
    
    //Efface la valeur de la cellule
    procedure InternalClearValue(const aIndex, aValue :byte);

  public
    //Grille
    property  Grid :TSudokuGrid read FGrid;

    //Dure de la dernire rsolution
    property Duration :extended read FDuration;

    //Nombre de SetValue
    property SetValueCount :integer   read FSetValueCount;

    //Renvoi la valeur contenue dans une cellule
    function GetValue(const aCol, aRow :byte) :byte;

    //Ecrit une valeur dans une cellule. Renvoi FALSE si valeur invalide.
    function  SetValue(const aCol, aRow, aValue :byte) :boolean; 

    //Efface la valeur de la cellule
    procedure ClearValue(const aCol, aRow :byte);

    //Efface la grille
    procedure Reset;

    //Here we go !!!
    function   Start :boolean;

    //No comment!
    constructor Create;
  end;

const
  //Constante utilise pour reseter la grille
  FullMask :word = $03FE; //0000001111111110

  //Tableau des 9 masques possibles. Est plus rapide que N fois shl/shr
  Masks : array[1..9] of word = ($0002,$0004,$0008,$0010,$0020,$0040,$0080,$0100,$0200);

//Compteur de prcision
function GetTickCountEx :Extended;

implementation

function GetTickCountEx :Extended;
var
  Frequency :Int64;
  Counter   :Int64;

begin
  //Si le compteur de prcision n'est pas disponible, utilise GetTickCount
  //converti en seconde.
  if QueryPerformanceFrequency(Frequency) then
  begin
    QueryPerformanceCounter(Counter);
    Result := Counter /Frequency;
  end
  else Result := GetTickCount *1000;
end;

{ TSudoku }

constructor TSudoku.Create;
var
  x, y :byte;
begin
  //Initialisation des donnes statiques des cellules
  for y := 0 to 8 do
    for x := 0 to 8 do
      with TSudokuGridRC(FGrid)[y, x] do
      begin
        Row    := y +1;
        Col    := x +1;
        Region := x div 3 +y div 3 *3 +1;
      end;

  Reset;
end;

procedure TSudoku.ClearValue(const aCol, aRow: byte);
begin
  with TSudokuGridRC(FGrid)[aRow, aCol] do
  begin
    Mask            := Mask or Masks[Value];
    Rows[Row]       := Rows[Row] or Masks[Value];
    Cols[Col]       := Cols[Col] or Masks[Value];
    Regions[Region] := Regions[Region] or Masks[Value];
    Value           := 0;
  end;
end;

function TSudoku.SetValue(const aCol, aRow, aValue: byte): boolean;
begin
  with TSudokuGridRC(FGrid)[aRow, aCol] do
  begin
    //Accept ?
    Result := Mask            and
              Rows[Row]       and
              Cols[Col]       and
              Regions[Region] and
              Masks[aValue] <> 0;

    if Result then
    begin
      //S'il y avait dj une valeur, l'efface
      if Value <> 0 then
        ClearValue(Col, Row);
        
      Value           := aValue;
      Mask            := Mask and not Masks[aValue];
      Rows[Row]       := Rows[Row] and not Masks[aValue];
      Cols[Col]       := Cols[Col] and not Masks[aValue];
      Regions[Region] := Regions[Region] and not Masks[aValue];
    end;
  end;
end;

function TSudoku.NextValue(const aIndex :byte; var aValue: byte): boolean;
begin
  //Cherche  partir de la dernire valeur teste (+1)
  while aValue <= 9 do
  begin
    with WorkGrid[aIndex]^ do
    begin
      if Mask               and
         Rows[Row]          and
         Cols[Col]          and
         Regions[Region]    and
         Masks[aValue] <> 0 then
      begin
        Result := TRUE;
        Exit;
      end
      else inc(aValue);
    end;
  end;

  Result := FALSE;
end;

function TSudoku.GetValue(const aCol, aRow :byte): byte;
begin
  Result := TSudokuGridRC(FGrid)[aRow, aCol].Value;
end;

procedure TSudoku.InternalClearValue(const aIndex, aValue: byte);
begin
  with WorkGrid[aIndex]^ do
  begin
    Value           := 0;
    Rows[Row]       := Rows[Row] or Masks[aValue];
    Cols[Col]       := Cols[Col] or Masks[aValue];
    Regions[Region] := Regions[Region] or Masks[aValue];
  end;
end;

procedure TSudoku.InternalSetValue(const aIndex, aValue: byte);
begin
  inc(FSetValueCount);

  with WorkGrid[aIndex]^ do
  begin
    Value           := aValue;
    Rows[Row]       := Rows[Row] and not Masks[aValue];
    Cols[Col]       := Cols[Col] and not Masks[aValue];
    Regions[Region] := Regions[Region] and not Masks[aValue];
  end;
end;

procedure TSudoku.Reset;
var
  i :byte;

begin
  for i := 1 to 81 do
    with FGrid[i] do
    begin
      Value := 0;
      Mask  := FullMask;
    end;

  for i := 1 to 9 do
  begin
    Rows[i]    := FullMask;
    Cols[i]    := FullMask;
    Regions[i] := FullMask;
  end;
end;

function TSudoku.Start :boolean;
var
  i     :byte;
  Count :byte;
  Value :byte;

begin
  Result := TRUE;
  FSetValueCount := 0;

  //Met l'application en priorit maximale (Pas sr que a serve !!!)
  SetPriorityClass(GetCurrentProcess, REALTIME_PRIORITY_CLASS);
  SetThreadPriority(GetCurrentThread, THREAD_PRIORITY_TIME_CRITICAL);

  FDuration := GetTickCountEx;

  Count := 0;

  //Cre la grille de travail. N'ajoute que les cellules  traiter
  for i := 1 to 81 do
    if FGrid[i].Value = 0 then
    begin
      inc(Count);
      WorkGrid[Count] := @FGrid[i];
    end;

  i     := 1;
  Value := 1;

  try
    //La rsolution proprement dite
    while i <= Count do
    begin
      if NextValue(i, Value) then
      begin
        InternalSetValue(i, Value);
        inc(i);
        Value := 1;
      end
      else
      begin
        Dec(i);
        Value := WorkGrid[i].Value +1;
        InternalClearValue(i, WorkGrid[i].Value);
      end;
    end;
  except
    Result := FALSE;
  end;

  FDuration := GetTickCountEx -FDuration;

  //Repasse l'application en priorit normale
  SetThreadPriority(GetCurrentThread, THREAD_PRIORITY_NORMAL);
  SetPriorityClass(GetCurrentProcess, NORMAL_PRIORITY_CLASS);
end;

end.

